import torch
import torch.nn as nn
import torch.nn.functional as F
import dgl
import networkx as nx
import numpy as np


torch.autograd.set_detect_anomaly(True)

class GATLayer(nn.Module):
    def __init__(self, in_feats, out_feats):
        super(GATLayer, self).__init__()
        self.fc = nn.Linear(in_feats, out_feats, bias=False)
        self.attn_fc = nn.Linear(2 * out_feats, 1, bias=False)

    def edge_attention(self, edges):
        z2 = torch.cat([edges.src['z'], edges.dst['z']], dim=1)
        a = self.attn_fc(z2)
        return {'e': F.leaky_relu(a)}

    def message_func(self, edges):
        return {'z': edges.src['z'], 'e': edges.data['e']}

    def reduce_func(self, nodes):
        alpha = F.softmax(nodes.mailbox['e'], dim=1)
        h = torch.sum(alpha * nodes.mailbox['z'], dim=1)
        return {'h': h}

    def forward(self, g, h):
        z = self.fc(h)
        g.ndata['z'] = z
        g.apply_edges(self.edge_attention)
        g.update_all(self.message_func, self.reduce_func)
        return g.ndata.pop('h')

class GAT(nn.Module):
    def __init__(self, in_feats, hidden_feats, out_feats, num_heads):
        super(GAT, self).__init__()
        self.layers = nn.ModuleList([
            GATLayer(in_feats, hidden_feats) if i == 0 else
            GATLayer(hidden_feats if i != num_heads - 1 else hidden_feats, out_feats)
            for i in range(num_heads)
        ])

    def forward(self, g, features, node_types):
        h = features
        for layer in self.layers[:-1]: 
            h = F.relu(layer(g, h))
        h = self.layers[-1](g, h)
        h = adjust_output_based_on_type(h, node_types)
        return h

